;;;; ss-to-html.l
;;;; From SakuraScript to HTML
;;;; Ver.1.0.3.20080623
;;;; 2005-12-16, 2008-06-23 d
;;;; Public Domain: Please feel free to reproduce or edit
;;;; http://purl.com/net/dkcc/shed/ss-to-html.html

;;;; "HTML" means ISO-HTML, of cource.

#|
CROWESSPt@X
http://crow.aqrs.jp/reference/all/

SakuraScript 2
http://jn.swee.to/s3c/sakurascript_2.html

f - XNvgt@X
http://futaba.sakura.st/sakurascript.html

[3 PDF
http://ssp.shillest.net/docs/ukgailly-novice.pdf

FrontPage - [
http://da.pekori.to/wiki/index.php?FrontPage
|#

#|
;;; SakuraScript (1)
\h\s[4]\p2\s[200]\u\s[10]iۂۂj\w9\w9\nH\w5lH\w9\nO̓efBA\w5ꉞ̂ڕtB\w9\w9\nȂɂł킩Ȃ̂ŁB\w9\w9\hc\w5c\w5cB\w9\w9\u\n\nȂ킯ł܂낵B\w9\w9\p2\s[204]Hׂ邩邩ǂɂȂB\w9\w9\h\n\n\s[21]c\w5c\w5c\w5cB\w9\ni\w5c\w5cBj\w9\w9\u\s[11]\n\nӂB\e
|#

;;; global functions

(defun replace-regexp-whole (regexp replace &optional case-fold)
  (interactive "*sregexp: \nsreplace: ")
  (save-excursion
    (goto-char (point-min))
    (replace-regexp (compile-regexp regexp case-fold) replace t)))

(defun html-convert-to-character-entity-reference ()
  (interactive "*")
  (replace-regexp-whole "&" "&amp;")
  (replace-regexp-whole "<" "&lt;")
  (replace-regexp-whole ">" "&gt;"))

(defmacro save-variables (var &body body)
  (let ((var-t (gensym)))
    `(let ((,var-t (mapcar #'eval ',var)))
       (unwind-protect
           (progn ,@body)
         (mapcar #'set ',var ,var-t)))))

;;; functions for preference

(defun ss-list-ghosts (&aux comp)
  (maphash #'(lambda (ghost names) (push ghost comp)) *ss-ghost-table*)
  comp)

(defun ss-set-ghost (&optional ghost)
  (interactive)
  (message "~A" (setq *ss-current-p-name* (gethash (if ghost ghost (completing-read "ghost: " (ss-list-ghosts) :must-match t)) *ss-ghost-table*))))

(defun ss-add-ghost (ghost &rest rest &aux name name-list (i 0))
  (while (setq name (pop rest))
    (setq name-list (acons i name name-list))
    (incf i))
  (when name-list
    (setf (gethash ghost *ss-ghost-table*) name-list)))

(defun ss-id-to-name (id)
  (cdr (assoc id *ss-current-p-name*)))

(defun ss-set-name-freely (&aux name (i 0) (old-names *ss-current-p-name*))
  (interactive)
  (setq *ss-current-p-name* ())
  (while (string/= (setq name (read-string (format nil "\\p~D: " i) :default (cdr (assoc i old-names)))) "")
    (setq *ss-current-p-name* (acons i name *ss-current-p-name*))
    (incf i))
  (message "~A" *ss-current-p-name*))

;;; preference (user can reset)

(defvar *ss-current-p-name* '((1 . "ɂイ") (0 . "")))

(defvar *ss-ghost-table* nil)
(unless *ss-ghost-table*
  (setq *ss-ghost-table* (make-hash-table :test 'equal))
  (ss-add-ghost "" "" "ɂイ")
  (ss-add-ghost "ot" "ot" "")
  (ss-add-ghost "Emily/Phase4" "G~" "efB" "G~I")
  (ss-add-ghost "˂ƃV[^qw2.2" "˂" "V[^")
  (ss-add-ghost "o(IB)" "o" "ɂイ")
  (ss-add-ghost "^}" "" "^}")
  (ss-add-ghost "NԂԊOn" "N" "q" "N" "q"))

(defvar *ss-insert-open-paragraph*
  (lambda (p p-name s)
    (insert (format nil "\t<p class=\"p~A\"><img alt=\"~A: \" src=\"surfaces/surface~4,'0D.png\">" p p-name s))))

(defvar *ss-insert-close-paragraph*
  (lambda (p p-name s)
    (insert "</p>\n")))

(defvar *ss-insert-open-dialogue*
  (lambda ()
    (insert "<div class=\"ss\">\n")))

(defvar *ss-insert-close-dialogue*
  (lambda ()
    (insert "</div>\n")))

(defvar *ss-insert-break-line*
  (lambda ()
    (insert "<br>\n\t")))

;;; core

(defmacro ss-read-char-from-region (char)
  `(setq ,char (following-char)))

(defmacro ss-pop-argument (arg)
  `(when (scan-buffer (compile-regexp "\\(.*?\\)\\]"))
     (setq ,arg (match-string 1))
     (delete-region (match-beginning 0) (match-end 0))))

(defmacro ss-parse-command (char &body body)
  `(when (setq ,char (following-char))
     (delete-char)
     (cond
      ,@body)))

(defmacro ss-close-paragraph (paragraph &optional p p-name s)
  `(when ,paragraph
     (funcall *ss-insert-close-paragraph* ,p ,p-name ,s)
     (setq ,paragraph ,p)))

(defun ss-to-html-core (&aux (p 0) (s #(0 10 200 300 400 500 600 700 800 900)) char str paragraph dialogue)
;  (interactive "*")
  (save-excursion
    (html-convert-to-character-entity-reference)
    (replace-regexp-whole "\n" "")
    (goto-char (point-min))
    (while (char/= (ss-read-char-from-region char) #\NUL)
      (if (char= char #\\) ;; \: command starts
          (progn
            (delete-char)
            (when (char/= (ss-read-char-from-region char) #\NUL)
              (when (char/= char #\\) ;; not \\: not "\"
                (delete-char)
                (cond
                 ((char= char #\s) ;; \s: surfaces
                  (ss-parse-command
                   char
                   ((char= char #\[) ;; \s[0], \s[1], ...
                    (ss-pop-argument str)
                    (when (setq str (parse-integer str :junk-allowed t))
                      (setf (svref s p) str)))
                   ((digit-char-p char) ;; \s0, \s1, ..., \s9
                    (setq char (digit-char-p char))
                    (when (= p 1) (incf char 10))
                    (setf (svref s p) char))))
                 ((char= char #\p) ;; \p: persona
                  (ss-parse-command
                   char
                   ((char= char #\[) ;; \p[0], \p[1], ...
                    (ss-pop-argument str)
                    (when (setq str (parse-integer str))
                      (setq p str)))
                   ((digit-char-p char) ;; \p0, \p1, ..., \p9
                    (setq p (digit-char-p char)))))
                 ((or (char= char #\h) (char= char #\0)) ;; \h, \0: Sakura
                  (setq p 0))
                 ((or (char= char #\u) (char= char #\1)) ;; \u, \1: Unyuu
                  (setq p 1))
                 ((char= char #\n) ;; \n: \n
                  (ss-parse-command
                   char
                   ((char= char #\[) ;; \n[half]
                    (ss-pop-argument str))
                   (t
                    (when (eq p paragraph)
                      (funcall *ss-insert-break-line*))
                    (insert char)
                    (backward-char))))
                 ((char= char #\w) ;; \w: wait
                  (ss-parse-command
                   char
                   ((digit-char-p char) nil)
                   (t (insert char))))
                 ((char= char #\e) ;; \e: end
                  (ss-close-paragraph paragraph p (ss-id-to-name p) (svref s p))
                  (funcall *ss-insert-close-dialogue*)
                  (setq dialogue nil))
                 ((char= char #\_) ;; \_: other command starts
                  (ss-parse-command
                   char
                   ((char= char #\w) ;; \_w: fine wait
                    (ss-parse-command
                     char
                     ((char= char #\[)
                      (ss-pop-argument str))
                     (t (insert char))))
                   (t (insert char)))))))) ;; \_?
        (progn ;; not \: normal text
          (if dialogue
              (progn
                (unless (eq p paragraph)
                  (funcall *ss-insert-close-paragraph* p (ss-id-to-name p) (svref s p))
                  (funcall *ss-insert-open-paragraph* p (ss-id-to-name p) (svref s p))
                  (setq paragraph p)))
            (progn
              (funcall *ss-insert-open-dialogue*)
              (setq dialogue t)
              (funcall *ss-insert-open-paragraph* p (ss-id-to-name p) (svref s p))
              (setq paragraph p)))
          (forward-char))))
    (goto-char (point-max))
    (ss-close-paragraph paragraph p (ss-id-to-name p) (svref s p))
    (when dialogue
      (funcall *ss-insert-close-dialogue*)
      (setq dialogue nil))))

;;; user presets (user can redefine)

(defmacro ss-preset (&body body)
  `(let ((bi (selected-buffer))
         (bi-min (point-min))
         (bi-max (point-max)))
     (switch-to-buffer "*SakuraScript*")
     (make-local-variable 'need-not-save)
     (setq need-not-save t)
     (erase-buffer (selected-buffer))
     (insert-buffer-substring bi bi-min bi-max)
     (save-variables
      (*ss-insert-open-paragraph*
       *ss-insert-close-paragraph*
       *ss-insert-open-dialogue*
       *ss-insert-close-dialogue*
       *ss-insert-break-line*)
      (progn ,@body)
      (ss-to-html-core))))

(defun ss-to-html-default ()
  (interactive "*")
  (ss-preset))

(defun ss-to-html ()
  (interactive "*")
  (ss-preset
   (setq *ss-insert-open-paragraph*
         (lambda (p p-name s)
             (insert (format nil "\t<p class=\"p~A\"><img alt=\"~A: \" src=\"surfaces/surface~4,'0D.png\">" p p-name s)))
         *ss-insert-close-paragraph*
         (lambda (p p-name s)
             (insert "</p>\n"))
         *ss-insert-open-dialogue*
         (lambda ()
             (insert "<div class=\"ss\">\n"))
         *ss-insert-close-dialogue*
         (lambda ()
             (insert "</div>\n"))
         *ss-insert-break-line*
         (lambda ()
             (insert "<br>\n\t")))))

(defun ss-to-xhtml ()
  (interactive "*")
  (ss-preset
   (setq *ss-insert-open-paragraph*
         (lambda (p p-name s)
             (insert (format nil "\t<p class=\"p~A\"><img alt=\"~A: \" src=\"surfaces/surface~4,'0D.png\" />" p p-name s)))
         *ss-insert-close-paragraph*
         (lambda (p p-name s)
             (insert "</p>\n"))
         *ss-insert-open-dialogue*
         (lambda ()
             (insert "<div class=\"ss\">\n"))
         *ss-insert-close-dialogue*
         (lambda ()
             (insert "</div>\n"))
         *ss-insert-break-line*
         (lambda ()
             (insert "<br />\n\t")))))

(defun ss-to-html-with-dl () ;; "NԂԊOn" format
  (interactive "*")
  (ss-preset
   (setq *ss-insert-open-paragraph*
         (lambda (p p-name s)
             (insert (format nil "<dt class=\"p~A\">~A</dt>\n" p p-name s))
             (insert (format nil "\t<dd class=\"p~A\">" p s)))
         *ss-insert-close-paragraph*
         (lambda (p p-name s)
             (insert "</dd>\n"))
         *ss-insert-open-dialogue*
         (lambda ()
             (insert "<dl class=\"ss\">\n"))
         *ss-insert-close-dialogue*
         (lambda ()
             (insert "</dl>\n"))
         *ss-insert-break-line*
         (lambda ()
             (insert "<br>\n\t")))))

(defun ss-to-html-without-img () ;; "܂" format
  (interactive "*")
  (ss-preset
   (setq *ss-insert-open-paragraph*
         (lambda (p p-name s)
             (insert (format nil "\t<p class=\"p~A s~D\"><em>~A:</em> " p s p-name)))
         *ss-insert-close-paragraph*
         (lambda (p p-name s)
             (insert "</p>\n"))
         *ss-insert-open-dialogue*
         (lambda ()
             (insert "<div class=\"ss\">\n"))
         *ss-insert-close-dialogue*
         (lambda ()
             (insert "</div>\n"))
         *ss-insert-break-line*
         (lambda ()
             (insert "<br>\n\t")))))

;;; playing

(defmacro ss-hide-the-rest (&body body)
  `(prog2
       (narrow-to-region (point-min) (point)) ;; hide the rest
       (progn ,@body)
     (widen)))

(defvar *ss-message-wait* 0.05)

(defun ss-play-ss (&aux (p 0) (s #(0 10 200 300 400 500 600 700 800 900)) char str paragraph dialogue)
  (interactive "*")
  (let ((bi (selected-buffer))
        (bi-min (point-min))
        (bi-max (point-max)))
    (switch-to-buffer "*SakuraScript*")
    (make-local-variable 'need-not-save)
    (setq need-not-save t)
    (erase-buffer (selected-buffer))
    (insert-buffer-substring bi bi-min bi-max)
    (save-variables
     (*ss-insert-open-paragraph*
      *ss-insert-close-paragraph*
      *ss-insert-open-dialogue*
      *ss-insert-close-dialogue*
      *ss-insert-break-line*
      hide-restricted-region)
     (setq *ss-insert-open-paragraph*
           (lambda (p p-name s)
               (insert (format nil "~A[~D]: " p-name s)))
           *ss-insert-close-paragraph*
           (lambda (p p-name s)
               (insert "\n"))
           *ss-insert-open-dialogue*
           (lambda ()
               (insert "Dialogue starts:\n")
               (ss-hide-the-rest (sit-for 1)))
           *ss-insert-close-dialogue*
           (lambda ()
               (insert "Dialogue ends.\n\n"))
           *ss-insert-break-line*
           (lambda ()
               (insert "\n    "))
           hide-restricted-region t)
     (save-excursion
       (save-restriction ;; for hide
         ;       (html-convert-to-character-entity-reference)
         (replace-regexp-whole "\n" "")
         (goto-char (point-min))
         (while (char/= (ss-read-char-from-region char) #\NUL)
           (if (char= char #\\) ;; \: command starts
               (progn
                 (delete-char)
                 (when (char/= (ss-read-char-from-region char) #\NUL)
                   (when (char/= char #\\) ;; not \\: not "\"
                     (delete-char)
                     (cond
                      ((char= char #\s) ;; \s: surfaces
                       (ss-parse-command
                        char
                        ((char= char #\[)
                         (ss-pop-argument str)
                         (when (setq str (parse-integer str :junk-allowed t))
                           (setf (svref s p) str)))
                        ((digit-char-p char) ;; \s0, \s1, ..., \s9
                         (setq char (digit-char-p char))
                         (when (= p 1) (incf char 10))
                         (setf (svref s p) char))))
                      ((char= char #\p) ;; \p: persona
                       (ss-parse-command
                        char
                        ((char= char #\[)
                         (ss-pop-argument str)
                         (when (setq str (parse-integer str))
                           (setq p str)))
                        ((digit-char-p char) ;; \p0, \p1, ..., \p9
                         (setq p (digit-char-p char)))))
                      ((or (char= char #\h) (char= char #\0)) ;; \h, \0: Sakura
                       (setq p 0))
                      ((or (char= char #\u) (char= char #\1)) ;; \u, \1: Unyuu
                       (setq p 1))
                      ((char= char #\n) ;; \n: \n
                       (ss-parse-command
                        char
                        ((char= char #\[) ;; \n[half]
                         (ss-pop-argument str))
                        (t
                         (when (eq p paragraph)
                           (funcall *ss-insert-break-line*))
                         (insert char)
                         (backward-char))))
                      ((char= char #\w) ;; \w: wait
                       (ss-parse-command
                        char
                        ((digit-char-p char) (ss-hide-the-rest (sit-for (* (digit-char-p char) 0.05)))) ;; wait
                        (t (insert char))))
                      ((char= char #\e) ;; \e: end
                       (ss-close-paragraph paragraph p (ss-id-to-name p) (svref s p))
                       (funcall *ss-insert-close-dialogue*)
                       (setq dialogue nil))
                      ((char= char #\_) ;; \_: other command starts
                       (ss-parse-command
                        char
                        ((char= char #\w) ;; \_w: fine wait
                         (ss-parse-command
                          char
                          ((char= char #\[)
                           (ss-pop-argument str)
                           (ss-hide-the-rest (sit-for (* (parse-integer str) 0.001)))) ;; wait
                          (t (insert char))))
                        (t (insert char)))))))) ;; \_?
             (progn ;; not \: normal text
               (if dialogue
                   (progn
                     (unless (eq p paragraph)
                       (funcall *ss-insert-close-paragraph* p (ss-id-to-name p) (svref s p))
                       (funcall *ss-insert-open-paragraph* p (ss-id-to-name p) (svref s p))
                       (setq paragraph p)))
                 (progn
                   (funcall *ss-insert-open-dialogue*)
                   (setq dialogue t)
                   (funcall *ss-insert-open-paragraph* p (ss-id-to-name p) (svref s p))
                   (setq paragraph p)))
               (ss-hide-the-rest (sit-for *ss-message-wait*))
               (forward-char))))
         (goto-char (point-max))
         (ss-close-paragraph paragraph p (ss-id-to-name p) (svref s p))
         (when dialogue
           (funcall *ss-insert-close-dialogue*)
           (setq dialogue nil))
         (ding))))))
